package plugins.big.bigsnakeutils.process.process2D;

import icy.image.IcyBufferedImage;
import icy.main.Icy;
import icy.sequence.Sequence;
import icy.type.DataType;
import plugins.big.bigsnakeutils.process.process1D.Convolver1D;
import plugins.big.bigsnakeutils.process.process3D.ImageArrayUtils;

public class Hessian2D {

	// ----------------------------------------------------------------------------
	// INPUT IMAGE

	/** Input image stack. */
	private final double[] image_;

	/** Width of the image stack to filter. */
	private final int width_;
	/** Height of the image stack to filter. */
	private final int height_;
	
	// ----------------------------------------------------------------------------
	// outPUT IMAGE

	/** score that indicates its probability to be a ridge (ellipsoidal shape and not a circle) */
	double[] merit_; 
	
	/** direction along x of the eigen vector linked to the minimum eigen value */
	double[] dx_;
	
	/** direction along y of the eigen vector linked to the minimum eigen value */
	double[] dy_;
	
	// ----------------------------------------------------------------------------
	// KERNELS
	
	/** Contains the kernel cst*exp(-(x^2)/(2*sigmaX^2)). */
	private final double kernelBaseX_[];
	
	/** Contains the kernel cst*exp(-(y^2)/(2*sigmaY^2)). */
	private final double kernelBaseY_[];
	
	/** Contains the kernel (-x/(sigmaX^2))*kernelBaseX_. */
	private final double kernelFact1X_[];
	
	/** Contains the kernel (-y/(sigmaY^2))*kernelBaseY_. */
	private final double kernelFact1Y_[];

	/** Contains the kernel ((x^2/sigmaX^4)-1/sigmaX^2)*kernelBaseX_. */
	private final double kernelFact2X_[];
	
	/** Contains the kernel ((y^2/sigmaY^4)-1/sigmaY^2)*kernelBaseY_. */
	private final double kernelFact2Y_[];
	
	// ----------------------------------------------------------------------------
	// CONSTANTS

	/**
	 * This number indicates the number of 2*sigmas that are considered in the
	 * filtering.
	 */
	private final static double GAUSSIAN_TRUNCATION = 3.0;
	
	// ============================================================================
	// PUBLIC METHODS

	/** Default constructor. */
	public Hessian2D(double[] array, int width, int height, double sigma) {
		
		image_ = new double[width * height];
		for (int i = 0; i < width * height; i++) {
			image_[i] = array[i];
		}
		

		width_ = width;
		height_ = height;

		kernelBaseX_ = createKernel_Base(sigma);
		kernelBaseY_ = createKernel_Base(sigma);
		
		kernelFact1X_ = createKernel_Fact1(sigma);
		kernelFact1Y_ = createKernel_Fact1(sigma);
		
		kernelFact2X_ = createKernel_Fact2(sigma);
		kernelFact2Y_ = createKernel_Fact2(sigma);
		
		merit_=new double[width_ * height_];	
		dx_=new double[width_ * height_];
		dy_=new double[width_ * height_];
		

	}
	
	public void filter() {

		/** h=f \ast (kernelX*kernelY)*/
		double[] hxx=filtering(kernelFact2X_,kernelBaseY_);
		double[] hyy=filtering(kernelBaseX_,kernelFact2Y_);
		double[] hxy=filtering(kernelFact1X_,kernelFact1Y_);
		
		//show(hxx, "hxx");
		//show(hyy, "hyy");
		//show(hxy, "hxy");
		double j=0;
		for (int i = 0; i < width_ * height_; i++) {
			double cxx=hxx[i];
			double cyy=hyy[i];
			double cxy=hxy[i];
			
			double delta = (cxx - cyy) * (cxx - cyy) + 4.0 * cxy * cxy;
			
			if (delta >= 0) {
				
				//Compute the eigen values
				delta = Math.sqrt(delta);
				double l1 = (cxx + cyy - delta) * 0.5;
				double l2 = (cxx + cyy + delta) * 0.5;
				
				//Black
//				double lmin=l2;//minimum eigen value
//				double lmax=l1;//maximum eigen value
//				
//				if(Math.abs(lmin)>Math.abs(lmax) && lmin>0){
//				//j++;
//				}
//				else{
//					lmin=0;
//					lmax=0;
//				}

				
				//White
				double lmin=l1;
				double lmax=l2;
				
				if(Math.abs(lmin)>Math.abs(lmax) && lmin<0){
				//j++;
				}
				else{
					lmin=0;
					lmax=0;
				}

				
				//Compute the merit
				double m = Math.sqrt(Math.abs(lmin) * Math.abs(lmax - lmin));//formule MU cours
				merit_[i]=m;
				
				//Compute the minimum eigen vector
				double vx = hxy[i];
				double vy = (lmin - hxx[i]);
				
				double norm = Math.sqrt(vx * vx + vy * vy);
				if(norm==0) {
					dx_[i]=0;
					dy_[i]=0;
				}else{
					dx_[i]=m*vx / norm;
					dy_[i]=m*vy / norm;
				}
			}
		}
		
		//show(merit_, "merit");
		//show(dx_, "dx");
		//show(dy_, "dy");
	
	}
	
	public double[] getMerit() {
		return merit_;
	}
	
	public double[] getdx() {
		return dx_;
	}
	
	public double[] getdy() {
		return dy_;
	}
	
	public void show(double[] im, String title){
		
		Sequence sequence = new Sequence();

		IcyBufferedImage tempImage = new IcyBufferedImage(width_, height_, 1, DataType.DOUBLE);
		
		double[] temp = tempImage.getDataXYAsDouble(0);
		for (int y = 0; y < height_; y++) {
		for (int x = 0; x < width_; x++) {

		temp[x + y * width_] = im[x + y * width_];

		}
		}
		tempImage.dataChanged();
		sequence.addImage(tempImage);
		sequence.setName(title);
		
		Icy.getMainInterface().addSequence(sequence);
	}
	
	// ============================================================================
	// PRIVATE METHODS
	
	/** Constructs the kernel cst*exp(-(t^2)/(2*sigma^2)). */
	private static double[] createKernel_Base(double sigma) {
		if (sigma <= 0.0) {
			double[] kernel = new double[1];
			kernel[0] = 1.0;
			return kernel;
		}

		double s2 = sigma * sigma;
		double dem = 2.0 * s2;
		int size = (int) Math
				.round(((int) (sigma * GAUSSIAN_TRUNCATION)) * 2.0 + 1);
		int size2 = size / 2;
		double[] kernel = new double[size];

		double x;
		for (int k = 0; k < size; k++) {
			x = (k - size2) * (k - size2);
			kernel[k] = Math.exp(-x / dem)/(sigma*Math.sqrt(2*Math.PI));
		}
		return kernel;
	}
	
	
	/** Constructs the kernel (-t/(sigma^2))*kernelBase). */
	private static double[] createKernel_Fact1(double sigma) {
		if (sigma <= 0.0) {
			double[] kernel = new double[1];
			kernel[0] = 1.0;
			return kernel;
		}

		double s2 = sigma * sigma;
		double dem = 2.0 * s2;
		int size = (int) Math
				.round(((int) (sigma * GAUSSIAN_TRUNCATION)) * 2.0 + 1);
		int size2 = size / 2;
		double[] kernel = new double[size];

		double x;
		for (int k = 0; k < size; k++) {
			x = (k - size2);
			kernel[k] = (- x / s2) * Math.exp(-x*x / dem)/(sigma*Math.sqrt(2*Math.PI));
		}
		return kernel;
	}
	
	/** Contains the kernel ((t^2/sigma^4)-1/sigma^2)*kernelBase. */
	private static double[] createKernel_Fact2(double sigma) {
		if (sigma <= 0.0) {
			double[] kernel = new double[1];
			kernel[0] = 1.0;
			return kernel;
		}

		double s2 = sigma * sigma;
		double s4 = s2 * s2;
		double dem = 2.0 * s2;
		int size = (int) Math
				.round(((int) (sigma * GAUSSIAN_TRUNCATION)) * 2.0 + 1);
		int size2 = size / 2;
		double[] kernel = new double[size];

		double x;
		for (int k = 0; k < size; k++) {
			x = (k - size2);
			kernel[k] = ((x*x/s4)-(1/s2)) * Math.exp(-x*x / dem)/(sigma*Math.sqrt(2*Math.PI));
		}
		return kernel;
	}
	
	private double[] filtering(double[] kernelX, double[] kernelY){
		double[][] output = new double[1][width_ * height_];
		
		for (int i = 0; i < width_ * height_; i++) {
			output[0][i] = image_[i];
		}


		double vinY[] = new double[height_];
		double voutY[] = new double[height_];
		for (int x = 0; x < width_; x++) {
			ImageArrayUtils.getY(output, width_, height_, 1, x,
					0, 0, vinY);
			Convolver1D.convolve(vinY, voutY, kernelY);
			ImageArrayUtils.putY(output, width_, height_, 1, x,
					0, 0, voutY);
		}

		double vinX[] = new double[width_];
		double voutX[] = new double[width_];
		for (int y = 0; y < height_; y++) {
			ImageArrayUtils.getX(output, width_, height_, 1, 0,
					y, 0, vinX);
			Convolver1D.convolve(vinX, voutX, kernelX);
			ImageArrayUtils.putX(output, width_, height_, 1, 0,
					y, 0, voutX);
		}
		
		return output[0];
	}


}
